#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

/*
  NOTE: for a real-mode DOS version this
  source code MUST be compiled under at least
  -ml (Large Model) memory model or you may hit the
  "not enough memory" errors on large .ART files
   with more than 256 tiles inside
*/

#ifndef _STDINT_H
typedef signed char    int8_t;
typedef unsigned char  uint8_t;
typedef short          int16_t;
typedef unsigned short uint16_t;
/* WIN */
#if INT_MAX == 0x7FFFFFFFL
typedef int            int32_t;
typedef unsigned int   uint32_t;
#ifndef _WIN32
#ifdef __WIN32
#define _WIN32 1
#else
#ifdef __WIN32__
#define _WIN32 1
#endif
#endif
#endif
#ifdef _WIN32
#define AM_VER "Win32"
#else
#define AM_VER "Arc32"
#endif
#endif
/* DOS */
#if INT_MAX == 0x7FFF
typedef long            int32_t;
typedef unsigned long  uint32_t;
#define AM_VER "DOS16"
#endif
#endif

/* some hacks to reduce win32 executable file size */
#ifdef _WIN32
#include "win32gcc.h"
#include "intwfile.h"
#endif

#define ART_VERSION 1L

#pragma pack(push, 1)
typedef struct {
  int32_t artversion;
  int32_t numtiles;
  int32_t localtilestart;
  int32_t localtileend;
} art_head;

typedef struct {
  int fl;
  art_head head;
  int32_t offs; /* v1.2 */
  int32_t size; /* v1.2 */
  int32_t data;
  int16_t *tilesizx;
  int16_t *tilesizy;
  int32_t *picanm;
} art_info;

typedef struct {
  char     name[12];
  uint32_t size;
} grp_item;

typedef struct {
  int fl;
  uint32_t size; /* total items */
} grp_info;
#pragma pack(pop)

void InitArtInfo(art_info *ai) {
  if (ai) {
    memset(ai, 0, sizeof(ai[0]));
    ai->fl = -1;
  }
}

void FreeArtFile(art_info *ai) {
  if (ai) {
    if (ai->picanm) { free(ai->picanm); }
    if (ai->tilesizy) { free(ai->tilesizy); }
    if (ai->tilesizx) { free(ai->tilesizx); }
    if (ai->fl != -1) { close(ai->fl); }
    InitArtInfo(ai);
  }
}

/*
  compare two strings
  c - NOT null terminated string, max 12 characters
  s - ASCIIZ null terminated string
*/
int MatchItem(char *c, char *s, int m) {
int i, r;
  r = 1;
  for (i = 0; i < 12; i++) {
    /* digits mask */
    if ((m) && (*s == '0')) {
      if ((*c >= '0') && (*c <= '9')) {
        c++;
        s++;
        continue;
      } else {
        r = 0;
        break;
      }
    }
    /* not equal chars - strings not matched */
    if (*c != *s) {
      r = 0;
      break;
    }
    /* string ends earlier */
    if (!*c) { break; }
    /* go to next chars */
    c++;
    s++;
  }
  /* if 12 chars exceeded
     and there are still
     chars left in s */
  if (*s) {
    r = 0;
  }
  return(r);
}

int OpenGroupFile(char *name) {
uint32_t i, sz, fsz;
grp_item item;
int fl;
  fl = -1;
  if (name) {
    fl = open(name, O_RDONLY | O_BINARY);
    if (fl != -1) {
      lseek(fl, 0, SEEK_END);
      fsz = tell(fl);
      lseek(fl, 0, SEEK_SET);
      sz = 0;
      /* sanity check */
      if (sizeof(item) <= fsz) {
        read(fl, &item, sizeof(item));
        /* some more sanity checks */
        if (MatchItem(item.name, "KenSilverman", 0) && (((item.size + 1) * sizeof(item)) <= fsz)) {
          /* sum all item sizes */
          sz = (item.size + 1) * sizeof(item);
          i = item.size;
          while (i--) {
            read(fl, &item, sizeof(item));
            /* even more sanity checks */
            if ((sz + item.size) > fsz) {
              fsz = 0;
              break;
            }
            sz += item.size;
          }
          /* whole file size check */
          sz = (sz <= fsz) ? 1 : 0;
        }
      }
      /* invalid group file */
      if (!sz) {
        close(fl);
        fl = -1;
      }
    }
  }
  return(fl);
}

int FindGroupTile(int fl, char *name, art_info *ai, uint32_t min_size) {
uint32_t i, sz;
grp_item item;
  sz = 0;
  if ((fl != -1) && name && ai && min_size) {
    ai->offs = 0;
    lseek(fl, 0, SEEK_SET);
    read(fl, &item, sizeof(item));
    sz = (item.size + 1) * sizeof(item);
    for (i = item.size; i; i--) {
      read(fl, &item, sizeof(item));
      /* found, few more sanity checks */
      if (MatchItem(item.name, name, 0) && (item.size >= min_size)) {
        /* item found */
        ai->offs = sz;
        ai->size = item.size;
        break;
      }
      /* calc absolute item offset */
      sz += item.size;
    }
    /* offset zero - item not found */
    if (!ai->offs) { sz = 0; }
  }
  return(sz ? fl : -1);
}

void LoadArtFile(char *name, art_info *ai, int fg) {
int32_t i, sz;
  sz = 0;
  InitArtInfo(ai);
  if (name && ai) {
    /* v1.2 */
    if (fg != -1) {
      ai->fl = FindGroupTile(fg, name, ai, sizeof(ai->head));
    } else {
      ai->fl = open(name, O_RDONLY | O_BINARY);
    }
    if (ai->fl != -1) {
      /* v1.2 */
      if (fg == -1) {
        /* get filesize */
        lseek(ai->fl, 0, SEEK_END);
        ai->offs = 0; /* v1.3 */
        ai->size = tell(ai->fl);
      }
      lseek(ai->fl, ai->offs, SEEK_SET);
      sz = ai->size;
      read(ai->fl, &ai->head, sizeof(ai->head));
      /* valid .ART file */
      ai->head.numtiles = ai->head.localtileend - ai->head.localtilestart + 1;
      ai->data = sizeof(ai->tilesizx[0]) + sizeof(ai->tilesizy[0]) + sizeof(ai->picanm[0]);
      ai->data = sizeof(ai->head) + (ai->head.numtiles * ai->data);
      if (
        (ai->head.artversion == ART_VERSION) && (ai->head.localtileend >= ai->head.localtilestart) &&
        (ai->head.localtileend >= 0) && (ai->head.localtilestart >= 0) && (sz >= ai->data)
      ) {
        /* allocate memory */
        ai->tilesizx = (int16_t *) malloc(ai->head.numtiles * sizeof(ai->tilesizx[0]));
        ai->tilesizy = (int16_t *) malloc(ai->head.numtiles * sizeof(ai->tilesizy[0]));
        ai->picanm   = (int32_t *) malloc(ai->head.numtiles * sizeof(ai->picanm[0]));
        /* memory allocated */
        if (ai->tilesizx && ai->tilesizy && ai->picanm) {
          /* read info */
          read(ai->fl, ai->tilesizx, ai->head.numtiles * sizeof(ai->tilesizx[0]));
          read(ai->fl, ai->tilesizy, ai->head.numtiles * sizeof(ai->tilesizy[0]));
          read(ai->fl, ai->picanm, ai->head.numtiles * sizeof(ai->picanm[0]));
          /* check file size */
          sz -= ai->data;
          for (i = 0; i < ai->head.numtiles; i++) {
            /* v1.1 */
            if ((ai->tilesizx[i] <= 0) || (ai->tilesizy[i] <= 0)) {
              ai->tilesizx[i] = 0;
              ai->tilesizy[i] = 0;
              ai->picanm[i] = 0;
            }
            sz -= ((int32_t) ai->tilesizx[i]) * ((int32_t) ai->tilesizy[i]);
          }
          if (sz >= 0) {
            /* done */
            sz = 1;
          } else {
            sz = 0;
            printf("\nError: corrupted Build .ART file or partial tile data.\n");
          }
        } else {
          sz = 0;
          printf("\nError: memory allocation error - not enough memory.\n");
        }
      } else {
        sz = 0;
        printf("\nError: not a valid Build .ART file or unsupported version.\n");
      }
    } else {
      printf("\nError: can not open input file for read.\n");
    }
  }
  /* any kind of error */
  if (!sz) {
    if (fg != -1) { ai->fl = -1; } /* v1.3 */
    FreeArtFile(ai);
  }
}

int32_t GetArtTileOffs(int32_t num, art_info *ai) {
int32_t i, r;
  r = 0;
  if ((ai) && (0 <= num) && (num < ai->head.numtiles)) {
    r = ai->data;
    for (i = 0; i < ai->head.numtiles; i++) {
      if (i == num) { break; }
      r += ((ai->tilesizx[i] > 0) && (ai->tilesizy[i] > 0)) ? (((int32_t) ai->tilesizx[i]) * ((int32_t) ai->tilesizy[i])) : 0;
    }
  }
  return(r);
}

/* static copy buffer */
#define BUF_MAX 0x4000
static uint8_t buf[BUF_MAX];
/* for CheckSameFile() */
#define BUF_LEN (BUF_MAX/2)

void CopyTileData(int fi, int fo, int32_t sz) {
int32_t l;
  if ((fi != -1) && (fo != -1) && (sz > 0)) {
    do {
      l = (sz < BUF_MAX) ? sz : BUF_MAX;
      sz -= l;
      read(fi, buf, l);
      write(fo, buf, l);
    } while (sz > 0);
  }
}

/*
  since in DOS you actually can delete opened file
  we need a way to somehow detect and avoid this case
  so if a second file exactly the same as first - don't touch it
*/
int CheckSameFile(art_info *sa, art_info *da) {
int32_t sz, l;
int r;
  /* by default - different file */
  r = 0;
  /* checks - from faster to slower ones */
  if (sa && da && (sa->fl != -1) && (da->fl != -1)) {
    /* compare header values */
    if ((!memcmp(&sa->head, &da->head, sizeof(sa->head))) && (sa->data == da->data)) {
      /* compare tiles info */
      if (
        (!memcmp(sa->tilesizx, da->tilesizx, sa->head.numtiles * sizeof(sa->tilesizx[0]))) &&
        (!memcmp(sa->tilesizy, da->tilesizy, sa->head.numtiles * sizeof(sa->tilesizy[0]))) &&
        (!memcmp(sa->picanm, da->picanm, sa->head.numtiles * sizeof(sa->picanm[0]))) &&
        (sa->size == da->size) /* v1.2 */
      ) {
        /* we are extremly unlucky - compare whole files */
        lseek(sa->fl, sa->offs + sa->data, SEEK_SET);
        lseek(da->fl, da->offs + da->data, SEEK_SET);
        sz = sa->size - sa->data;
        do {
          l = (sz < BUF_LEN) ? sz : BUF_LEN;
          sz -= l;
          read(sa->fl, buf, l);
          read(da->fl, &buf[BUF_LEN], l);
          /* let's hope there differences somewhere early in the file */
          r = memcmp(buf, &buf[BUF_LEN], BUF_LEN) ? 0 : 1;
        } while ((sz > 0) && r);
        /* v1.2 */
        lseek(sa->fl, sa->offs, SEEK_SET);
        lseek(da->fl, da->offs, SEEK_SET);
      }
    }
  }
  return(r);
}

/* v1.2 */
int CheckUpdateTiles(art_info *sa, art_info *da) {
int r;
uint32_t i;
  r = 0;
  /* sanity check */
  if (sa && da) {
    /* range check */
    if (
      (da->head.localtilestart <= sa->head.localtileend) &&
      (da->head.localtileend >= sa->head.localtilestart)
    ) {
      /* check non-empty tiles */
      for (i = sa->head.localtilestart; i <= sa->head.localtileend; i++) {
        if (i < da->head.localtilestart) { continue; }
        if (i > da->head.localtileend) { break; }
        /* at least one non-empty file */
        if (sa->tilesizx[i - sa->head.localtilestart]) {
          r = 1;
          break;
        }
      }
    }
  }
  return(r);
}

void MakeTileName(char *s, int n) {
  if (s) { sprintf(s, "TILES%03d.ART", n); }
}

static char stTemp[] = "TILES$$$.TMP";

int MakeTileList(uint8_t *list, int fl) {
grp_item item;
uint32_t i;
char s[13];
int k;
  k = 0;
  if (list) {
    memset(list, 0, 125);
    if (fl == -1) {
      /* search in current folder */
      for (i = 0; i < 1000; i++) {
        MakeTileName(s, i);
        if (access(s, 0) != -1) {
          list[i >> 3] |= 1 << (i & 0x07);
        }
      }
    } else {
      /* search in group file */
      MakeTileName(s, 0);
      lseek(fl, 0, SEEK_SET);
      read(fl, &item, sizeof(item));
      i = item.size;
      while (i--) {
        read(fl, &item, sizeof(item));
        /* got tile file */
        if (MatchItem(item.name, s, 1)) {
          k = ((item.name[5] - '0') * 100) + ((item.name[6] - '0') * 10) + (item.name[7] - '0');
          list[k >> 3] |= 1 << (k & 0x07);
        }
      }
    }
    /* got something? */
    i = 0;
    for (k = 0; k < 125; k++) {
      if (list[k]) {
        i = 1;
        break;
      }
    }
    k = i ? 1 : 0;
  }
  return(k);
}

int main(int argc, char *argv[]) {
/* (125 * 8) = 1000 [000...999] */
uint8_t tilemask[125];
art_info sa, da;
int32_t i, k, n, sz;
int fl, t, fg;
char s[13];
  printf("Build .ART tile merger v1.3 ["AM_VER"]\n(c) -=CHE@TER=- 2017,2018,2020\nhttp://ctpax-cheater.losthost.org/\n\n");
  if ((argc < 2) || (argc > 3)) {
    printf(
      "Usage: artmerge <filename.art> [filename.grp]\n\n"
      "Where:\n"
      "  filename.art - tiles .ART file with the tiles to update (empty skipped)\n"
      "  filename.grp - optional .GRP group file (DUKE3D.GRP, STUFF.DAT, etc.)\n\n"
      "If the group file was specified then all modified TILES###.ART files will be\n"
      "extracted otherwise this tool will search in current folder for these files.\n"
      "\n"
    );
    return(1);
  }
  /* v1.3 */
  fg = -1;
  if (argc == 3) {
    fg = OpenGroupFile(argv[2]);
    if (fg == -1) {
      printf("Error: invalid group file specified or can not open for read.\n\n");
      return(2);
    }
  }
  printf("* Building available TILES###.ART list\n");
  if (!MakeTileList(tilemask, fg)) {
    if (fg != -1) { close(fg); }
    printf("Error: no tile files TILES###.ART found.\n\n");
    return(3);
  }
  printf("* Merging from %s\n", argv[1]);
  LoadArtFile(argv[1], &sa, -1);
  /* file loaded */
  if (sa.fl != -1) {
    printf("\n");
    /* for each .ART file */
    for (t = 0; t < 1000; t++) {
      /* if tile file not exist */
      if (!(tilemask[t >> 3] & (1 << (t & 0x07)))) { continue; }
      MakeTileName(s, t);
      printf("%s\x0D", s);
      LoadArtFile(s, &da, fg);
      if (da.fl != -1) {
        /* will use as flags later */
        sz = 1;
        fl = -1;
        /* check that this file in range */
        if (
          (!CheckSameFile(&sa, &da)) &&
          /* v1.2 */
          (CheckUpdateTiles(&sa, &da))
        ) {
          sz = 0;
          printf("\n");
          /* temporary file */
          fl = open(stTemp, O_RDWR | O_BINARY | O_CREAT | O_TRUNC, S_IREAD | S_IWRITE);
          if (fl != -1) {
            /* set header as stub for now */
            write(fl, &da.head, sizeof(da.head));
            write(fl, da.tilesizx, da.head.numtiles * sizeof(da.tilesizx[0]));
            write(fl, da.tilesizy, da.head.numtiles * sizeof(da.tilesizy[0]));
            write(fl, da.picanm, da.head.numtiles * sizeof(da.picanm[0]));
            /* data size */
            sz = da.data;
            lseek(da.fl, da.offs + sz, SEEK_SET); /* v1.1 fix */
            /* create new tiles file */
            for (i = 0; i < da.head.numtiles; i++) {
              k = (i + da.head.localtilestart);
              n = (k - sa.head.localtilestart);
              n = ((int32_t) sa.tilesizx[n]) * ((int32_t) sa.tilesizy[n]);
              /* tile exists in source file and not empty */
              if ((sa.head.localtilestart <= k) && (k <= sa.head.localtileend) && (n)) {
                k -= sa.head.localtilestart;
                /* update tile info */
                lseek(da.fl, ((int32_t) da.tilesizx[i]) * ((int32_t) da.tilesizy[i]), SEEK_CUR); /* v1.1 fix */
                da.tilesizx[i] = sa.tilesizx[k];
                da.tilesizy[i] = sa.tilesizy[k];
                da.picanm[i] = sa.picanm[k];
                lseek(sa.fl, GetArtTileOffs(k, &sa), SEEK_SET);
                CopyTileData(sa.fl, fl, n);
                sz += n;
              } else {
                /* copy old tile if not empty */
                n = ((int32_t) da.tilesizx[i]) * ((int32_t) da.tilesizy[i]);
                CopyTileData(da.fl, fl, n);
                sz += n;
              }
            }
            /* check that all data correctly saved */
            sz = (sz == tell(fl)) ? 1 : 0;
            /* write fixed header */
            if (sz) {
              lseek(fl, 0, SEEK_SET);
              write(fl, &da.head, sizeof(da.head));
              write(fl, da.tilesizx, da.head.numtiles * sizeof(da.tilesizx[0]));
              write(fl, da.tilesizy, da.head.numtiles * sizeof(da.tilesizy[0]));
              write(fl, da.picanm, da.head.numtiles * sizeof(da.picanm[0]));
            }
            close(fl);
          }
        }
        if (fg != -1) { da.fl = -1; } /* v1.3 */
        FreeArtFile(&da);
        /* everything ok */
        if (sz) {
          if (fl != -1) {
            /* remove old */
            remove(s);
            /* replace with temp file */
            rename(stTemp, s);
          }
        } else {
          if (fl == -1) {
            printf("Error: can not create temporary file.\n");
          } else {
            /* remove temp file */
            remove(stTemp);
            printf("Error: temporary file partially saved, probably not enough free disk space.\n");
          }
          /* v1.3 */
          *s = 0;
          break;
        }
      }
    }
    /* clear output */
    printf("%12c\n", ' ');
    if (*s) { /* v1.3 */
      printf("done\n\n");
    }
    FreeArtFile(&sa);
  }
  /* v1.3 */
  if (fg != -1) { close(fg); }
  return(0);
}
